/* animationengine.c	-- Yasuhiro Yamazaki Sat Jan  5 2002
 */

#include <stdio.h>

#include "modelfile.h"
#include "keyframefile.h"
#include "keyframebind.h"
#include "animationengine.h"

/* create */
AnimationEngine *
createAnimationEngine( void )
{
  AnimationEngine *ret = (AnimationEngine *)malloc( sizeof( AnimationEngine ) );
  ret->ref = 1;
  ret->bind = NULL;
  ret->model = NULL;
  return ret;
}

/* destroy */
void 
destroyAnimationEngine( AnimationEngine *a )
{
  if( a!=NULL ){
    destroyKeyFrameBinds( a->bind );
    destroyModelFile( a->model );
    free( a );
  }
}

/* set params */
static KeyFrameBind *
constructKeyFrameBindFromModelBoneStructure( ModelBone *root )
{
  KeyFrameBind *ret = NULL, *add;
  if( root!=NULL ){
    ret = createKeyFrameBind( root->name );
    setModelBoneKeyFrameBind( ret, root );

    add = constructKeyFrameBindFromModelBoneStructure( root->child );
    if( add!=NULL ){
      add->ref++;
      linkKeyFrameBind( ret, add );
      destroyKeyFrameBind( add );
    }
    add = constructKeyFrameBindFromModelBoneStructure( root->brother );
    if( add!=NULL ){
      add->ref++;
      linkKeyFrameBind( ret, add );
      destroyKeyFrameBind( add );
    }
  }
  return ret;
}
void 
setAnimationModelToAE( AnimationEngine *a, ModelFile *m )
{
  if( a!=NULL ){
    destroyKeyFrameBinds( a->bind );
    destroyModelFile( a->model );

    a->model = m;
    a->bind = constructKeyFrameBindFromModelBoneStructure( getModelBoneFromModelFile( m ) );
  }
}

void 
setNewAnimationToAE( AnimationEngine *a, KeyFrameFile *k, const double current_time )
{
  if( a!=NULL && k!=NULL ){
    KeyFrameBind *bind = a->bind;
    KeyFrame *key = getKeyFrameFromKeyFrameFile( k );
    while( bind!=NULL ){
      addKeyFrameKeyFrameBind( bind, key, current_time );
      bind = nextKeyFrameBind( bind );
    }
    destroyKeyFrame( key );
  }
}

void 
executeAnimationEngine( AnimationEngine *a, const double current_time )
{
  if( a!=NULL ){
    KeyFrameBind *bind = a->bind;
    while( bind!=NULL ){
      applyKeyFrameBind( bind, current_time );
      expireOldKeyFrameKeyFrameBind( bind, current_time );
      bind = nextKeyFrameBind( bind );
    }
  }
}

