/* collisiontriangle.c	-- Yasuhiro Yamazaki Sat Jan  5 2002
 */
#include <stdio.h>
#include <stdlib.h>

#include "vector3d.h"
#include "collisionface.h"
#include "collisiontriangle.h"

#define EPSILON 0.01f

/* åȤΰִط򸡺 */
static double
checkEdge( CollisionTriangle *t, Vector3D *intersection, const int from, const int to )
{
  Vector3D v0, v1, v2;
  subVector3D( &v0, &t->vertex[to], &t->vertex[from] );
  subVector3D( &v1, intersection, &t->vertex[from] );
  crossProductVector3D( &v2, &v0, &v1 );
  return (double)dotProductVector3D( &t->normal, &v2 );
}

/* η׻ */
static int
detectCollisionTriangleWithVector3D( CollisionTriangle *t, Vector3D *from, Vector3D *to,
                                     Vector3D *intersection, Vector3D *direction )
{
  int ret = 0;
  Vector3D pass;
  double dot_product;
  subVector3D( &pass, to, from );
  
  /* 1st. stage 
  * Check a possibility of pass through on the face */
  if( (dot_product=dotProductVector3D( &t->normal, from ) - t->d) *
      (            dotProductVector3D( &t->normal,   to ) - t->d)<0.0f ){
    double s0, s1;
    /* 2nd. stage 
    * Calculate a point of intersection with the vector on the face */
    double length = (t->d - dotProductVector3D( &t->normal, from ))/dotProductVector3D( &t->normal, &pass );
    
    /* find an intersection point which is expressed by "pass * length + from" */
    multiplyVector3D( &pass, (float)length );
    addVector3D( intersection, &pass, from );
    
    /* 3rd. stage
    * Check intersection point is inside or outside of all edges */
    s0 = checkEdge( t, intersection, 0, 1 );
    s1 = checkEdge( t, intersection, 1, 2 );
    if( !(s0*s1<0.0) && !(s0*checkEdge( t, intersection, 2, 0 )<0.0) ){
      /* pass all: DETECTED! */
      ret = 1;
      copyVector3D( direction, &t->normal );
      if( dot_product<0.0 ){
        multiplyVector3D( direction, -1.0f );
      }
    }
  }
  return ret;
}

/* Ƚ (٥ȥ̤ˡ˦Ť餵줿) */
static int
detectAlgo1CollisionTriangleWithVector3D( void *itself, Vector3D *from, Vector3D *to, 
                                          Vector3D *modified_to )
{
  int ret = 0;

  if( itself!=NULL && from!=NULL && to!=NULL && modified_to!=NULL ){
    CollisionTriangle *t = (CollisionTriangle *)itself;
    Vector3D intersection, direction, tmp;

    if( (ret=detectCollisionTriangleWithVector3D( t, from, to, &intersection, &direction )) ){
      /* vector modification */
      copyVector3D( &tmp, &direction );
      multiplyVector3D( &tmp, EPSILON );
      addVector3D( modified_to, &intersection, &tmp );
    }
  }
  return ret;
}

/* Ƚ (٥ȥϥ٥ȥν̾ޤǰư˦Ť餵줿 ) */
static int
detectAlgo2CollisionTriangleWithVector3D( void *itself, Vector3D *from, Vector3D *to, 
                                          Vector3D *modified_to )
{
  int ret = 0;

  if( itself!=NULL && from!=NULL && to!=NULL && modified_to!=NULL ){
    CollisionTriangle *t = (CollisionTriangle *)itself;
    Vector3D intersection, direction, tmp;

    if( (ret=detectCollisionTriangleWithVector3D( t, from, to, &intersection, &direction )) ){
      /* vector modification */
      double n = (t->d - dotProductVector3D( &t->normal, to ))/dotProductVector3D(&t->normal, &direction) + EPSILON;
      copyVector3D( &tmp, &direction );
      multiplyVector3D( &tmp, (float)n );
      addVector3D( modified_to, to, &tmp );
    }
  }
  return ret;
}

/*  */
CollisionTriangle *
createCollisionTriangle( const int algo )
{
  CollisionTriangle *ret = (CollisionTriangle *)malloc( sizeof( CollisionTriangle ) );
  initializeDefaultCollisionFace( &ret->common );
  switch( algo ){
  case COLLISION_TRIANGLE_ALGO_DEFAULT:
    ret->common.detectionWithVector3D = detectAlgo1CollisionTriangleWithVector3D;
    break;
  case COLLISION_TRIANGLE_ALGO_VECTOR_SMOOTH:
    ret->common.detectionWithVector3D = detectAlgo2CollisionTriangleWithVector3D;
    break;
  }
  setVector3DCoordinatef( &ret->vertex[0], 0.0f, 0.0f, 0.0f );
  setVector3DCoordinatef( &ret->vertex[1], 0.0f, 0.0f, 0.0f );
  setVector3DCoordinatef( &ret->vertex[2], 0.0f, 0.0f, 0.0f );
  return ret;
}

/* ˴ */
void 
destroyCollisionTriangle( CollisionTriangle *c )
{
  if( c!=NULL ){
    if( --c->common.ref<=0 ){
      free( c );
    }
  }
}

/* ѷ򥻥å */
void 
setTriangleCollisionTriangle( CollisionTriangle *target, 
                              Vector3D *v0, Vector3D *v1, Vector3D *v2 )
{
  Vector3D a, b;

  if( v0!=NULL && v1!=NULL && v2!=NULL ){
    copyVector3D( &target->vertex[0], v0 );
    copyVector3D( &target->vertex[1], v1 );
    copyVector3D( &target->vertex[2], v2 );

    /* caluclation of face information of triangle */
    subVector3D( &a, v1, v0 );
    subVector3D( &b, v2, v0 );
    crossProductVector3D( &target->normal, &a, &b );
    normalizeVector3D( &target->normal );

    target->d = dotProductVector3D( &target->normal, v0 );
  }
}

/* ץؿ */
int 
detectWithVector3DCollisionTriangle( CollisionTriangle *target, Vector3D *from, Vector3D *to, 
                                     Vector3D *modified_to )
{
  int ret = 0;
  if( target!=NULL ){
    /* Force transformation of pointer to common interface */
    CollisionFace *i = (CollisionFace *)target;
    ret = i->detectionWithVector3D( i, from, to, modified_to );
  }
  return ret;
}
