/* drop.c	-- Yasuhiro Yamazaki Tue Nov  6 2001
 */
#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "vector3d.h"
#include "collisionface.h"
#include "collisiontriangle.h"

/* ѷο */
#define NUMBER_OF_TRIANGLES 64

/* ѷΥ () */
#define TRIANGLE_SIZE       0.75f

/* γҤο */
#define NUMBER_OF_PARTICLES 128

/*  */
#define GRAVITY             0.001f

/* «ϰ */
#define BOUNDING_AREA_X_MIN (-2.0f)
#define BOUNDING_AREA_Y_MIN (-3.0f)
#define BOUNDING_AREA_Z_MIN (-2.0f)
#define BOUNDING_AREA_X_MAX ( 2.0f)
#define BOUNDING_AREA_Y_MAX ( 3.0f)
#define BOUNDING_AREA_Z_MAX ( 2.0f)

/* ޥ */
#define RND(X) (((double)rand()/(double)RAND_MAX)*(double)(X))

/* ɥξ֤򼨤 */
#define WINDOW_STATE_SHOW 0
#define WINDOW_STATE_HIDE 1

/* ž ɥʬɥåȤɤž뤫򼨤 */
#define ROTATION_COEFFICIENT 180.0 /* degree */

/* ΰư 
   ɥʬɥåȤɤ餬ư뤫򼨤 */
#define DISTANCE_COEFFICIENT   5.0 /* distance */

/* γҤι¤ */
typedef struct tagParticle
{
  int ref;
  Vector3D position, velocity;
} Particle;

/* Static ѿ */
static int window_state = WINDOW_STATE_SHOW;   /* ɥξ */
static double ratio_x = 0.0, ratio_y = 0.0;    /* ޥɸη׻ݻ */
static float rotate_x = 0.0f, rotate_y = 0.0f; /* ɽǥβž */
static float distance = 7.5f;                  /*  */
static int push_button;                        /* 줿ܥݻ */

static CollisionTriangle *triangle[NUMBER_OF_TRIANGLES];
static Particle *particle[NUMBER_OF_PARTICLES];

/* «줿Υ */
static double
getRandomPoint( const double value_min, const double value_max )
{
  return RND(value_max-TRIANGLE_SIZE-(value_min+TRIANGLE_SIZE))+value_min+TRIANGLE_SIZE;
}
/* «줿Υʥ٥ȥ */
static void
randomVector( Vector3D *v, const double x, const double y, const double z )
{
  setVector3DCoordinatef( v, 
                          x + RND( TRIANGLE_SIZE*2.0 )-TRIANGLE_SIZE,
                          y + RND( TRIANGLE_SIZE*2.0 )-TRIANGLE_SIZE,
                          z + RND( TRIANGLE_SIZE*2.0 )-TRIANGLE_SIZE );
}
/* «줿Υʻѷ */
static void
randomTriangle( Vector3D *v0, Vector3D *v1, Vector3D *v2 )
{
  double x, y, z;
  x = getRandomPoint( BOUNDING_AREA_X_MIN, BOUNDING_AREA_X_MAX );
  y = getRandomPoint( BOUNDING_AREA_Y_MIN, BOUNDING_AREA_Y_MAX );
  z = getRandomPoint( BOUNDING_AREA_Z_MIN, BOUNDING_AREA_Z_MAX );

  randomVector( v0, x, y, z );
  randomVector( v1, x, y, z );
  randomVector( v2, x, y, z );
}

/* γҤν */
static void
initializeParticle( Particle *p )
{
  setVector3DCoordinatef( &p->position, 
                          getRandomPoint( BOUNDING_AREA_X_MIN, BOUNDING_AREA_X_MAX ),
                          BOUNDING_AREA_Y_MAX,
                          getRandomPoint( BOUNDING_AREA_Z_MIN, BOUNDING_AREA_Z_MAX ) );
  setVector3DCoordinatef( &p->velocity, 0.0f, 0.0f, 0.0f );
}
/* γҤ */
static Particle *
createParticle( void )
{
  Particle *ret = (Particle *)malloc( sizeof( Particle ) );
  ret->ref = 1;
  initializeParticle( ret );
  return ret;
}
/* γҤ˴ */
static void
destroyParticle( Particle *p )
{
  if( p!=NULL && --p->ref<=0 ){
    free( p );
  }
}

/* ʻѷν */
static void
initializeTriangles( void )
{
  int i;
  Vector3D v0, v1, v2;

  for( i=0; i<NUMBER_OF_TRIANGLES; i++ ){
    triangle[i] = createCollisionTriangle( COLLISION_TRIANGLE_ALGO_VECTOR_SMOOTH );

    randomTriangle( &v0, &v1, &v2 );
    setTriangleCollisionTriangle( triangle[i], &v0, &v1, &v2 );
  }

}
/* ʻѷ˴ */
static void
freeTriangles( void )
{
  int i;
  for( i=0; i<NUMBER_OF_TRIANGLES; i++ ){
    destroyCollisionTriangle( triangle[i] );
  }
}

/* Ƥν */
static void
initializeAll( void )
{
  int i;

  initializeTriangles();

  for( i=0; i<NUMBER_OF_PARTICLES; i++ ){
    particle[i] = createParticle();
  }
}

/* Ƥθ */
static void
freeAll( void )
{
  int i;

  freeTriangles();
  for( i=0; i<NUMBER_OF_PARTICLES; i++ ){
    destroyParticle( particle[i] );
  }
}

/*  */
static void
processAll( void )
{
  int i, j, again;
  Vector3D gravity, to, modified;
  setVector3DCoordinatef( &gravity, 0.0f, -GRAVITY, 0.0f );

  /* γҤ */
  for( i=0; i<NUMBER_OF_PARTICLES; i++ ){
    Particle *p = particle[i];

    addVector3D( &to, &p->position, &p->velocity );
    
    /* ͤ򸡺 */
    do{
      again = 0;
      for( j=0; j<NUMBER_OF_TRIANGLES; j++ ){
        CollisionFace *c = (CollisionFace *)triangle[j];
        if( c->detectionWithVector3D( c, &p->position, &to, &modified ) ){
          again++;
          subVector3D( &p->velocity, &modified, &p->position );
          addVector3D( &to, &p->position, &p->velocity );
        }
      }
    } while( again>0 && again<NUMBER_OF_TRIANGLES );

    /* γҤΰư */
    addVector3D( &p->position, &p->position, &p->velocity );
    addVector3D( &p->velocity, &p->velocity, &gravity );
    if( p->position.y<BOUNDING_AREA_Y_MIN ){
      initializeParticle( p );
    }
  }
}

/* ɽ */
static void 
putSomething( void )
{
  GLfloat color_red[4] = {1.0f, 0.0f, 0.0f, 0.0f};
  int i, j;

  /* ѷ */
  glEnable( GL_LIGHTING );

  glMaterialf ( GL_FRONT, GL_SHININESS,   128.0f );
  glMaterialfv( GL_FRONT, GL_DIFFUSE,  color_red );
  glMaterialfv( GL_FRONT, GL_SPECULAR, color_red );
  glMaterialfv( GL_FRONT, GL_AMBIENT,  color_red );

  glBegin( GL_TRIANGLES );
  for( i=0; i<NUMBER_OF_TRIANGLES; i++ ){
    glNormal3f( triangle[i]->normal.x, triangle[i]->normal.y, triangle[i]->normal.z );
    for( j=0; j<3; j++ ){
      glVertex3f( triangle[i]->vertex[j].x, triangle[i]->vertex[j].y, triangle[i]->vertex[j].z );
    }
  }
  glEnd();

  glDisable( GL_LIGHTING );
  
  /* γҤ */
  glColor3f( 0.0f, 1.0f, 1.0f );
  glPointSize( 2.0f );
  glBegin( GL_POINTS );
  for( i=0; i<NUMBER_OF_PARTICLES; i++ ){
    glVertex3f( particle[i]->position.x, particle[i]->position.y, particle[i]->position.z ); 
  }
  glEnd();

  /* ŷ */
  glColor3f( 0.0f, 0.0f, 1.0f );
  glBegin( GL_LINES );
  glVertex3f( BOUNDING_AREA_X_MIN, BOUNDING_AREA_Y_MAX, BOUNDING_AREA_Z_MIN );
  glVertex3f( BOUNDING_AREA_X_MAX, BOUNDING_AREA_Y_MAX, BOUNDING_AREA_Z_MIN );

  glVertex3f( BOUNDING_AREA_X_MIN, BOUNDING_AREA_Y_MAX, BOUNDING_AREA_Z_MIN );
  glVertex3f( BOUNDING_AREA_X_MIN, BOUNDING_AREA_Y_MAX, BOUNDING_AREA_Z_MAX );

  glVertex3f( BOUNDING_AREA_X_MAX, BOUNDING_AREA_Y_MAX, BOUNDING_AREA_Z_MIN );
  glVertex3f( BOUNDING_AREA_X_MAX, BOUNDING_AREA_Y_MAX, BOUNDING_AREA_Z_MAX );
  
  glVertex3f( BOUNDING_AREA_X_MIN, BOUNDING_AREA_Y_MAX, BOUNDING_AREA_Z_MAX );
  glVertex3f( BOUNDING_AREA_X_MAX, BOUNDING_AREA_Y_MAX, BOUNDING_AREA_Z_MAX );
  glEnd();
}

/* 褬ɬפȤʤä˸ƤФ */
static void
displayCallback( void )
{
  GLfloat light_position[] = { 0.0f, 0.0f, 1.0f, 0.0f };
  GLfloat light_one[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  
  glLoadIdentity();

  /*  */
  gluLookAt( 0.0, 0.0, distance, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0 );

  /* ХåեΥꥢ */
  glClear( GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT );

  /* 饤 */
  glLightfv( GL_LIGHT0, GL_POSITION, light_position );
  glLightfv( GL_LIGHT0, GL_DIFFUSE, light_one );
  glLightfv( GL_LIGHT0, GL_SPECULAR, light_one );
  glEnable( GL_LIGHTING );
  glEnable( GL_LIGHT0 );

  glPushMatrix();
    glRotatef( rotate_y, 0.0f, 1.0f, 0.0f);
    glRotatef( rotate_x, 1.0f, 0.0f, 0.0f);
    /* Put something */
    putSomething();
  glPopMatrix();

  /* ХåեΥå */
  glutSwapBuffers();
}

/* ɥꥵ줿˸ƤФ */
static void
reshapeCallback( int width, int height )
{
  glViewport( 0, 0, width, height );

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  gluPerspective( 60.0, width/(GLdouble)height, 0.01, 100 );

  glMatrixMode( GL_MODELVIEW );
}

/* ɥϾƤФ */
static void
idleCallback( void )
{
  if( window_state==WINDOW_STATE_SHOW ){
    /* ɥ˹Ԥ򤳤˵ */
    processAll();

    /* ɬפС򤹤 */
    glutPostRedisplay();
  }
}

/* ɥ{ɽ/줿}˸ƤФ */
static void
visibilityCallback( int state )
{
  switch( state ){
  case GLUT_NOT_VISIBLE:
    window_state = WINDOW_STATE_HIDE;
    break;
  case GLUT_VISIBLE:
    window_state = WINDOW_STATE_SHOW;
    break;
  }
}

/* ޥɸΥɥФ׻ */
static void
calculationRatioOfMouseCoordinate( const int x, const int y )
{
  int width, height;
  width  = glutGet( GLUT_WINDOW_WIDTH );
  height = glutGet( GLUT_WINDOW_HEIGHT );
  ratio_x = (double)y/(double)height;
  ratio_y = (double)x/(double)width;
}

/* ޥΥܥѲä˸ƤФ */
static void
mouseCallback( int button, int state, int x, int y )
{
  switch( button ){
  case GLUT_LEFT_BUTTON:
    push_button = GLUT_LEFT_BUTTON;
    break;
  case GLUT_MIDDLE_BUTTON:
    push_button = GLUT_MIDDLE_BUTTON;
    break;
  case GLUT_RIGHT_BUTTON:
    push_button = GLUT_RIGHT_BUTTON;
    break;
  }
  
  switch( state ){
  case GLUT_DOWN:
    break;
  case GLUT_UP:
    break;
  }

  calculationRatioOfMouseCoordinate( x, y );
}

/* ޥΥݥ󥿤ưä˸ƤФ */
static void
passiveMotionCallback( int x, int y )
{
}

/* ɥå桢ޥΥݥ󥿤ưä˸ƤФ */
static void
motionCallback( int x, int y )
{
  double rx, ry;

  switch( push_button ){ /* Ƥܥ˱ȿ򤵤 */
  case GLUT_LEFT_BUTTON:
    /* ʪΤβž׻ */
    rx = ratio_x;
    ry = ratio_y;
    calculationRatioOfMouseCoordinate( x, y );
    rotate_x += (float)((ratio_x-rx)*ROTATION_COEFFICIENT);
    rotate_y += (float)((ratio_y-ry)*ROTATION_COEFFICIENT);
    glutPostRedisplay();
    break;
  case GLUT_MIDDLE_BUTTON:
    /* ȥεΥѲ */
    rx = ratio_x;
    ry = ratio_y;
    calculationRatioOfMouseCoordinate( x, y );
    distance += ((float)(ratio_x-rx)+(float)(ratio_y-ry))*DISTANCE_COEFFICIENT;
    glutPostRedisplay();
    break;
  case GLUT_RIGHT_BUTTON:
    break;
  }
}

/* Ϥä˸ƤФ */
static void
keyboardCallback( unsigned char key_code, int x, int y )
{
  switch( key_code ){
  /*
  ** Esc줿ץλ
  */
  case 27:
    exit( EXIT_SUCCESS );
    break;

  case 32: /* space */
    freeTriangles();
    initializeTriangles();
    break;

  default:
    break;
  }
}

/* OpenGL ν */
static void
initGL( void )
{
  int width, height;

  width  = glutGet( GLUT_WINDOW_WIDTH );
  height = glutGet( GLUT_WINDOW_HEIGHT );

  reshapeCallback( width, height );

  glClearColor( 0.0, 0.0, 0.0, 1.0 );
  glEnable( GL_DEPTH_TEST );
  glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
}

/* MAIN */
int
main( int argc, char *argv[] )
{
  glutInit( &argc, argv );
  glutInitDisplayMode( GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH );

  glutCreateWindow( "Demo: Particles with Collision" );

  glutDisplayFunc      ( displayCallback       );
  glutKeyboardFunc     ( keyboardCallback      );
  glutReshapeFunc      ( reshapeCallback       );
  glutIdleFunc         ( idleCallback          );
  glutVisibilityFunc   ( visibilityCallback    );
  glutMouseFunc        ( mouseCallback         );
  glutPassiveMotionFunc( passiveMotionCallback );
  glutMotionFunc       ( motionCallback        );
  
  initGL(); /* OpenGL ν */
  initializeAll(); /*  */
  
  glutMainLoop(); /* ᥤ롼פγ */

  freeAll(); /*  */
  
  return EXIT_SUCCESS;
}
