/* keyframefile.c	-- Yasuhiro Yamazaki Sat Jan  5 2002
 */
#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>

#include "utility.h"
#include "keyframe.h"
#include "keyframefile.h"

static KeyFrameFile *
createKeyFrameFile( void )
{
  KeyFrameFile *ret = (KeyFrameFile *)malloc( sizeof( KeyFrameFile ) );
  ret->ref = 1;
  ret->filename = NULL;
  ret->key_frame = NULL;
  return ret;
}

static void
setFilenameKeyFrameFile( KeyFrameFile *k, const char *fn )
{
  if( k!=NULL && fn!=NULL ){
    if( k->filename!=NULL ){
      free( k->filename );
    }
    k->filename = (char *)malloc( (strlen( fn )+1) * sizeof( char ) );
    strcpy( k->filename, fn );
  }
}

static KeyFrame *
constructKeyFrameFromStream( FILE *read )
{
  char tmp[256];
  KeyFrame *begin = NULL;

  KeyFrame *state = NULL;
  double key_timing = 0.0;
  GLuint id = 0;
  
  while( !feof( read ) ){
    if( fscanf( read, "%s", tmp )!=EOF ){
      stringToLower( tmp );

      if( *tmp=='#' ){
	/* comment by '\n' */
	while( !feof( read ) && fgetc( read )!='\n' );
	
      } else if( strcmp( tmp, "key" )==0 ){
        /* set state of timing */
	fscanf( read, "%s", tmp );
        key_timing = atof( tmp );
        if( state!=NULL ){
          destroyKeyFrame( state );
        }
        state = createKeyFrame( key_timing );
        if( begin==NULL ){
          begin = state;
          begin->ref++;
        } else {
          linkKeyFrame( begin, state );
        }

      } else if( strcmp( tmp, "id" )==0 ){
	/* set state of ID Number */
	fscanf( read, "%d", &id );
	
      } else if( strcmp( tmp, "rotate1" )==0 ){
	/* add key for rotate1 */
        KeyValueAbstract *v = createKeyValue( KEY_VALUE_TYPE_ROTATE1, id );
	fscanf( read, "%lf", &v->rotate1.degree );
        addKeyValueToKeyFrame( state, v );
        destroyKeyValue( v );
	
      } else if( strcmp( tmp, "rotate2" )==0 ){
	/* add key for rotate2 */
        KeyValueAbstract *v = createKeyValue( KEY_VALUE_TYPE_ROTATE2, id );
	fscanf( read, "%lf", &v->rotate2.degree );
        addKeyValueToKeyFrame( state, v );
        destroyKeyValue( v );
	
      } else {
	/* ERROR */
	fprintf( stderr, "Syntax error \"%s\" in line. (Key=%lf / ID=%d)\n", 
                 tmp, key_timing, id );
      }
    }
  }
  
  return begin;
}

KeyFrameFile *
loadKeyFrameFile( const char *filename )
{
  FILE *read;
  KeyFrameFile *k = createKeyFrameFile();
  setFilenameKeyFrameFile( k, filename );
  read = fopen( filename, "r" );
  if( read!=NULL ){
    k->key_frame = constructKeyFrameFromStream( read );
    fclose( read );
  }
  return k;
}

void 
destroyKeyFrameFile( KeyFrameFile *k )
{
  if( k!=NULL ){
    if( --k->ref<=0 ){
      if( k->filename!=NULL ){
        free( k->filename );
      }
      if( k->key_frame!=NULL ){
        destroyKeyFrame( k->key_frame );
      }
      free( k );
    }
  }
}

KeyFrame *
getKeyFrameFromKeyFrameFile( KeyFrameFile *k )
{
  if( k!=NULL ){
    if( k->key_frame!=NULL ){
      k->key_frame->ref++;
    }
    return k->key_frame;
  }
  return NULL;
}
