/* keyframefile.c	-- Yasuhiro Yamazaki Sat Jan  5 2002
 */
#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>

#include "keyframe.h"
#include "keyframelist.h"

static unsigned int unique_no = 0;

KeyFrameList *
createKeyFrameList( void )
{
  KeyFrameList *ret = (KeyFrameList *)malloc( sizeof( KeyFrameList ) );
  ret->ref = 1;
  ret->no = unique_no++;
  ret->key_frame = NULL;
  ret->next = NULL;
  return ret;
}

void 
destroyKeyFrameList( KeyFrameList *k )
{
  if( k!=NULL ){
    if( --k->ref<=0 ){
      free( k );
    }
  }
}

void 
destroyKeyFrameLists( KeyFrameList *begin )
{
  KeyFrameList *target = begin, *reserve;
  while( target!=NULL ){
    reserve = target;
    target = target->next;
    destroyKeyFrameList( reserve );
  }
}

void 
linkKeyFrameList( KeyFrameList *target, KeyFrameList *branch )
{
  KeyFrameList *last;
  
  if( target!=NULL ){
    last = target;
    while( last->next!=NULL ){
      last = last->next;
    }
    last->next = branch;

    if( branch!=NULL ){
      branch->ref++;
    }
  }
}

KeyFrame *
getKeyFrameFromKeyFrameList( KeyFrameList *k )
{
  if( k!=NULL ){
    if( k->key_frame!=NULL ){
      k->key_frame->ref++;
    }
    return k->key_frame;
  }
  return NULL;
}

void
setKeyFrameToKeyFrameList( KeyFrameList *target, KeyFrame *k )
{
  if( target!=NULL ){
    destroyKeyFrame( target->key_frame );
    target->key_frame = k;
    if( k!=NULL ){
      k->ref++;
    }
  }
}

KeyFrame *
searchKeyFrameFromKeyFrameLists( KeyFrameList *begin, const unsigned int no )
{
  KeyFrameList *target = begin;
  while( target!=NULL ){
    if( target->no==no ) break;
    target = target->next;
  }
  if( target!=NULL && target->key_frame!=NULL ){
    target->key_frame->ref++;
    return target->key_frame;
  }
  return NULL;
}

/* enum */
KeyFrameList *
nextKeyFrameList( KeyFrameList *k )
{
  KeyFrameList *next = NULL;
  if( k!=NULL && k->next!=NULL ){
    next = k->next;
  }
  return next;
}

/* delete item (return new list pointer) */
KeyFrameList *
deleteKeyFrameList( KeyFrameList *begin, const GLuint id )
{
  KeyFrameList *ret = begin, *target = begin, *reserve = NULL;
  while( target!=NULL ){
    if( target->key_frame==NULL || 
        target->key_frame->list==NULL || 
        target->key_frame->list->common.id==id ){
      // delete item
      if( reserve==NULL ){
        ret = target->next;
      } else {
        reserve->next = target->next;
      }
      destroyKeyFrames( target->key_frame );
      destroyKeyFrameList( target );
    } else {
      reserve = target;
    }
    target = target->next;
  }
  return ret;
}

