/* keyvalue.c	-- Yasuhiro Yamazaki Sat Jan  5 2002
 */
#include <stdio.h>
#include <stdlib.h>

#include "vector3d.h"
#include "keyvalue.h"

/* ִؿ */
static void
linearInterpolationRotate1( void *dst, void *src1, void *src2, double rate )
{
  double neg_rate;
  KeyValueRotate1 *d = (KeyValueRotate1 *)dst;
  KeyValueAbstract *a;
  KeyValueRotate1 *s1 = (KeyValueRotate1 *)src1, *s2 = (KeyValueRotate1 *)src2;

  /* ݥ󥿤θ */
  if( dst==NULL || src1==NULL || src2==NULL ) return;

  /* ¤Τθ */
  a = (KeyValueAbstract *)src1;
  if( a->common.type!=KEY_VALUE_TYPE_ROTATE1 ) return;

  a = (KeyValueAbstract *)src2;
  if( a->common.type!=KEY_VALUE_TYPE_ROTATE1 ) return;
  
  /*  */
  neg_rate = 1.0 - rate;
  d->degree = s1->degree * neg_rate + s2->degree * rate;
}
static void
addRotate1( void *dst, void *src )
{
  KeyValueAbstract *d = (KeyValueAbstract *)dst, *s = (KeyValueAbstract *)src;
  if( dst==NULL || d->common.type!=KEY_VALUE_TYPE_ROTATE1 ||
      src==NULL || s->common.type!=KEY_VALUE_TYPE_ROTATE1 ) return;
  d->rotate1.degree += s->rotate1.degree;
}

static void
linearInterpolationRotate2( void *dst, void *src1, void *src2, double rate )
{
  double neg_rate;
  KeyValueRotate2 *d = (KeyValueRotate2 *)dst;
  KeyValueAbstract *a;
  KeyValueRotate2 *s1 = (KeyValueRotate2 *)src1, *s2 = (KeyValueRotate2 *)src2;

  /* ݥ󥿤θ */
  if( dst==NULL || src1==NULL || src2==NULL ) return;

  /* ¤Τθ */
  a = (KeyValueAbstract *)src1;
  if( a->common.type!=KEY_VALUE_TYPE_ROTATE2 ) return;

  a = (KeyValueAbstract *)src2;
  if( a->common.type!=KEY_VALUE_TYPE_ROTATE2 ) return;
  
  /*  */
  neg_rate = 1.0 - rate;
  d->degree = s1->degree * neg_rate + s2->degree * rate;
}
static void
addRotate2( void *dst, void *src )
{
  KeyValueAbstract *d = (KeyValueAbstract *)dst, *s = (KeyValueAbstract *)src;
  if( dst==NULL || d->common.type!=KEY_VALUE_TYPE_ROTATE2 ||
      src==NULL || s->common.type!=KEY_VALUE_TYPE_ROTATE2 ) return;
  d->rotate2.degree += s->rotate2.degree;
}

/* ֥å */
static void
clearInterface( KeyValueInterface *k )
{
  k->ref = 1;
  k->type = KEY_VALUE_TYPE_NONE;
  k->next = NULL;
  k->id   = 0;
  k->linearInterpolation = NULL;
  k->add = NULL;
}

static KeyValueAbstract *
createKeyValueNone( void )
{
  KeyValueAbstract *ret = (KeyValueAbstract *)malloc( sizeof( KeyValueAbstract ) );
  KeyValueNone *k = &ret->none;
  clearInterface( &k->common );
  return ret;
}

static KeyValueAbstract *
createKeyValueRotate1( void )
{
  KeyValueAbstract *ret = (KeyValueAbstract *)malloc( sizeof( KeyValueAbstract ) );
  KeyValueRotate1 *k = &ret->rotate1;
  clearInterface( &k->common );
  k->common.linearInterpolation = linearInterpolationRotate1;
  k->common.add = addRotate1;
  k->degree = 0.0;
  return ret;
}

static KeyValueAbstract *
createKeyValueRotate2( void )
{
  KeyValueAbstract *ret = (KeyValueAbstract *)malloc( sizeof( KeyValueAbstract ) );
  KeyValueRotate2 *k = &ret->rotate2;
  clearInterface( &k->common );
  k->common.linearInterpolation = linearInterpolationRotate2;
  k->common.add = addRotate2;
  k->degree = 0.0;
  return ret;
}

/*  */
KeyValueAbstract *
createKeyValue( const int type, const int id )
{
  KeyValueAbstract *ret = NULL;
  switch( type ){
  case KEY_VALUE_TYPE_NONE:
    ret = createKeyValueNone();
    break;
  case KEY_VALUE_TYPE_ROTATE1:
    ret = createKeyValueRotate1();
    break;
  case KEY_VALUE_TYPE_ROTATE2:
    ret = createKeyValueRotate2();
    break;
  }
  if( ret!=NULL ){
    ret->common.type = type;
    ret->common.id   = id;
  }
  return (KeyValueAbstract *)ret;
}

/* ˴ */
void 
destroyKeyValue( KeyValueAbstract *k )
{
  if( k!=NULL ){
    if( --k->common.ref<=0 ){
      free( k );
    }
  }
}
void 
destroyKeyValues( KeyValueAbstract *begin )
{
  KeyValueAbstract *target = begin, *reserve;
  while( target!=NULL ){
    reserve = target;
    target = target->common.next;
    destroyKeyValue( reserve );
  }
}

/* ʣ */
/* ʳǤ˥Υ򤹤äƤʤΤǤб */
KeyValueAbstract *
croneKeyValue( KeyValueAbstract *k )
{
  KeyValueAbstract *ret = (KeyValueAbstract *)malloc( sizeof( KeyValueAbstract ) );
  memcpy( ret, k, sizeof( KeyValueAbstract ) );
  ret->common.ref = 1;
  return ret;
}

int 
copyKeyValue( KeyValueAbstract *dst, KeyValueAbstract *src )
{
  int ref = dst->common.ref;
  memcpy( dst, src, sizeof( KeyValueAbstract ) );
  dst->common.ref = ref;
  return 0;
}

/* ³ */
void
linkKeyValue( KeyValueAbstract *target, KeyValueAbstract *branch )
{
  KeyValueAbstract *last;
  
  if( target!=NULL ){
    /* 򸡺 */
    last = target;
    while( last->common.next!=NULL ){
      last = last->common.next;
    }
    last->common.next = branch;
    if( branch!=NULL ){
      branch->common.ref++;
    }
  }
}

/* enum */
KeyValueAbstract *
nextKeyValue( KeyValueAbstract *k )
{
  KeyValueAbstract *next = NULL;
  if( k!=NULL ){
    next = k->common.next;
  }
  return next;
}

