/* modelbone.c	-- Yasuhiro Yamazaki Mon Dec 24 2001
 */
#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>

#include <GL/gl.h>

#include "vector3d.h"
#include "trimodel.h"
#include "modelbone.h"

/* static ѿ */
static GLuint global_auto_id = 0;

/*  */
ModelBone *
createModelBone( const char *model_filename )
{
  ModelBone *m = (ModelBone *)malloc( sizeof( ModelBone ) );
  m->ref = 1;
  m->name = ++global_auto_id;
  setVector3DCoordinatef( &m->scale,     1.0f, 1.0f, 1.0f );
  setVector3DCoordinatef( &m->translate, 0.0f, 0.0f, 0.0f );
  setVector3DCoordinatef( &m->center,    0.0f, 0.0f, 0.0f );
  setVector3DCoordinatef( &m->axis1,     1.0f, 0.0f, 0.0f );
  setVector3DCoordinatef( &m->axis2,     0.0f, 0.0f, 1.0f );
  m->rotate1 = 0.0;
  m->rotate2 = 0.0;
  m->model = loadTriModel( model_filename );
  m->brother = NULL;
  m->child   = NULL;
  return m;
}

/* ˴ */
void 
destroyModelBone( ModelBone *m )
{
  if( m!=NULL ){
    if( --m->ref<=0 ){
      if( m->model!=NULL ){
        destroyTriModel( m->model );
        m->model = NULL;
      }
      free( m );
    }
  }
}

void 
destroyLinkedModelBone( ModelBone *root )
{
  if( root!=NULL ){
    destroyLinkedModelBone( root->child );
    destroyLinkedModelBone( root->brother );
    destroyModelBone( root );
  }
}

/*  */
void 
linkAsBrotherModelBone( ModelBone *root, ModelBone *brother )
{
  if( root!=NULL && brother!=NULL ){
    /* ûҤõƥ󥯤 */
    while( root->brother!=NULL ){
      root = root->brother;
    }
    root->brother = brother;
    brother->ref++;
  }
}
void 
linkAsChildModelBone( ModelBone *parent, ModelBone *child )
{
  if( parent!=NULL && child!=NULL ){
    if( parent->child!=NULL ){
      /* Ƥ˻Ҥ顢λҤηȤƥ󥯤 */
      ModelBone *brother = parent->child;
      linkAsBrotherModelBone( brother, child );
    } else {
      /* ƤФơƤλҤʤ餽Τޤޥ󥯤 */
      parent->child = child;
      if( child!=NULL ){
        child->ref++;
      }
    }
  }
}

/* ɽ */
void 
putModelBone( const ModelBone *m )
{
  if( m!=NULL ){
    glPushMatrix();
    {
      glTranslatef( m->translate.x, m->translate.y, m->translate.z );
      glRotatef( (float)m->rotate1, m->axis1.x, m->axis1.y, m->axis1.z );
      glRotatef( (float)m->rotate2, m->axis2.x, m->axis2.y, m->axis2.z );
      glTranslatef( -m->center.x, -m->center.y, -m->center.z );
      glScalef( m->scale.x, m->scale.y, m->scale.z );
      putTriModel( m->model );
    }
    glPopMatrix();
  }
}

void 
putLinkedModelBone( const ModelBone *root )
{
  if( root!=NULL ){
    glPushMatrix();
    {
      glTranslatef( root->translate.x, root->translate.y, root->translate.z );
      glRotatef( (float)root->rotate1, root->axis1.x, root->axis1.y, root->axis1.z );
      glRotatef( (float)root->rotate2, root->axis2.x, root->axis2.y, root->axis2.z );
      glTranslatef( -root->center.x, -root->center.y, -root->center.z );
      glPushMatrix();
      {
	glScalef( root->scale.x, root->scale.y, root->scale.z );
	putTriModel( root->model );
      }
      glPopMatrix();
      putLinkedModelBone( root->child ); /* Ҷɽ */
    }
    glPopMatrix();
    putLinkedModelBone( root->brother ); /* ɽ */
  }
}

/* ѥ᡼ѹ */
void
setIDNumberModelBone( ModelBone *m, const GLuint id )
{
  if( m!=NULL ){
    m->name = id;
  }
}
void 
setScaleModelBone( ModelBone *m, const Vector3D scale )
{
  if( m!=NULL ){
    copyVector3D( &m->scale, &scale );
  }
}
void 
setTranslateModelBone( ModelBone *m, const Vector3D translate )
{
  if( m!=NULL ){
    copyVector3D( &m->translate, &translate );
  }
}
void 
setCenterModelBone( ModelBone *m, const Vector3D center )
{
  if( m!=NULL ){
    copyVector3D( &m->center, &center );
  }
}
void 
setAxis1ModelBone( ModelBone *m, const Vector3D axis1 )
{
  if( m!=NULL ){
    copyVector3D( &m->axis1, &axis1 );
  }
}
void 
setAxis2ModelBone( ModelBone *m, const Vector3D axis2 )
{
  if( m!=NULL ){
    copyVector3D( &m->axis2, &axis2 );
  }
}
void 
setRotate1ModelBone( ModelBone *m, const double rotate1 )
{
  if( m!=NULL ){
    m->rotate1 = rotate1;
  }
}
void 
setRotate2ModelBone( ModelBone *m, const double rotate2 )
{
  if( m!=NULL ){
    m->rotate2 = rotate2;
  }
}

/* ID 򥵡 */
ModelBone *
searchLinkedModelBone( ModelBone *root, const GLuint id )
{
  ModelBone *ret = NULL;
  if( root!=NULL ){
    if( root->name==id ) return root;
    if( (ret = searchLinkedModelBone( root->child, id ))==NULL ){
      ret = searchLinkedModelBone( root->brother, id );
    }
  }
  return ret;
}
