/* simplecollision.c	-- Yasuhiro Yamazaki Tue Nov  6 2001
 */
#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "collisionface.h"
#include "collisiontriangle.h"

/* ɥξ֤򼨤 */
#define WINDOW_STATE_SHOW 0
#define WINDOW_STATE_HIDE 1

/* ž ɥʬɥåȤɤž뤫򼨤 */
#define ROTATION_COEFFICIENT 180.0 /* degree */

/* ΰư 
   ɥʬɥåȤɤ餬ư뤫򼨤 */
#define DISTANCE_COEFFICIENT   5.0 /* distance */

/* Static ѿ */
static int window_state = WINDOW_STATE_SHOW;   /* ɥξ */
static double ratio_x = 0.0, ratio_y = 0.0;    /* ޥɸη׻ݻ */
static float rotate_x = 0.0f, rotate_y = 0.0f; /* ɽǥβž */
static float distance = 5.0f;                  /*  */
static int push_button;                        /* 줿ܥݻ */

static float x = 0.0f, y = 0.0f, vx = 0.0f, vy = 0.0f;
static CollisionTriangle *triangle = NULL;

/*  */
static void
initializeAll( void )
{
  triangle = createCollisionTriangle( COLLISION_TRIANGLE_ALGO_DEFAULT );
}

/*  */
static void
freeAll( void )
{
  destroyCollisionTriangle( triangle );
}

/* ɽ */
static void 
putSomething( void )
{
  Vector3D v0, v1, v2, from, to;
  GLfloat color_red[4] = {1.0f, 0.0f, 0.0f, 0.0f};
  int i;

  CollisionFace *c = (CollisionFace *)triangle;

  /* ɽ */

  /* ѷ */
  setVector3DCoordinatef( &v0,  0.0f, -2.0f, -1.0f );
  setVector3DCoordinatef( &v1,  1.0f,  1.0f,  1.0f );
  setVector3DCoordinatef( &v2, -2.0f, -1.0f,  0.0f );
  setTriangleCollisionTriangle( triangle, &v0, &v1, &v2 );

  /* ٥ȥ */
  setVector3DCoordinatef( &from,  1.0f+x, -2.0f+y,  2.0f );
  setVector3DCoordinatef( &to,   -1.0f+x,  2.0f+y, -2.0f );

  /* ͤθ */
  c->detectionWithVector3D( c, &from, &to, &to );

  /*  */
  glMaterialf ( GL_FRONT, GL_SHININESS,    32.0f );
  glMaterialfv( GL_FRONT, GL_DIFFUSE,  color_red );
  glMaterialfv( GL_FRONT, GL_SPECULAR, color_red );
  glMaterialfv( GL_FRONT, GL_AMBIENT,  color_red );

  glBegin( GL_TRIANGLES );
  glNormal3f( triangle->normal.x, triangle->normal.y, triangle->normal.z );
  for( i=0; i<3; i++ ){
    glVertex3f( triangle->vertex[i].x, triangle->vertex[i].y, triangle->vertex[i].z );
  }
  glEnd();

  glDisable( GL_LIGHTING );
  glColor3f( 0.0f, 1.0f, 1.0f );
  glBegin( GL_LINES );
  glVertex3f( from.x, from.y, from.z );
  glVertex3f(   to.x,   to.y,   to.z );
  glEnd();
  
  glEnable( GL_LIGHTING );
}

/* 褬ɬפȤʤä˸ƤФ */
static void
displayCallback( void )
{
  GLfloat light_position[] = { 0.0f, 0.0f, 1.0f, 0.0f };
  GLfloat light_one[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  
  glLoadIdentity();

  /*  */
  gluLookAt( 0.0, 0.0, distance, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0 );

  /* ХåեΥꥢ */
  glClear( GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT );

  /* 饤 */
  glLightfv( GL_LIGHT0, GL_POSITION, light_position );
  glLightfv( GL_LIGHT0, GL_DIFFUSE, light_one );
  glLightfv( GL_LIGHT0, GL_SPECULAR, light_one );
  glEnable( GL_LIGHTING );
  glEnable( GL_LIGHT0 );

  glPushMatrix();
    glRotatef( rotate_y, 0.0f, 1.0f, 0.0f);
    glRotatef( rotate_x, 1.0f, 0.0f, 0.0f);
    /* Put something */
    putSomething();
  glPopMatrix();

  /* ХåեΥå */
  glutSwapBuffers();
}

/* ɥꥵ줿˸ƤФ */
static void
reshapeCallback( int width, int height )
{
  glViewport( 0, 0, width, height );

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  gluPerspective( 60.0, width/(GLdouble)height, 0.01, 100 );

  glMatrixMode( GL_MODELVIEW );
}

/* ɥϾƤФ */
static void
idleCallback( void )
{
  if( window_state==WINDOW_STATE_SHOW ){
    /* ɥ˹Ԥ򤳤˵ */
    x += vx;
    y += vy;

    if( vx<0.0f ){
      vx += 0.001f;
      if( vx>0.0f ){
        vx = 0.0f;
      }
    } else if( vx>0.0f ){
      vx -= 0.001f;
      if( vx<0.0f ){
        vx = 0.0f;
      }
    }
    
    if( vy<0.0f ){
      vy += 0.001f;
      if( vy>0.0f ){
        vy = 0.0f;
      }
    } else if( vy>0.0f ){
      vy -= 0.001f;
      if( vy<0.0f ){
        vy = 0.0f;
      }
    }
    
    /* ɬפС򤹤 */
    glutPostRedisplay();
  }
}

/* ɥ{ɽ/줿}˸ƤФ */
static void
visibilityCallback( int state )
{
  switch( state ){
  case GLUT_NOT_VISIBLE:
    window_state = WINDOW_STATE_HIDE;
    break;
  case GLUT_VISIBLE:
    window_state = WINDOW_STATE_SHOW;
    break;
  }
}

/* ޥɸΥɥФ׻ */
static void
calculationRatioOfMouseCoordinate( const int x, const int y )
{
  int width, height;
  width  = glutGet( GLUT_WINDOW_WIDTH );
  height = glutGet( GLUT_WINDOW_HEIGHT );
  ratio_x = (double)y/(double)height;
  ratio_y = (double)x/(double)width;
}

/* ޥΥܥѲä˸ƤФ */
static void
mouseCallback( int button, int state, int x, int y )
{
  switch( button ){
  case GLUT_LEFT_BUTTON:
    push_button = GLUT_LEFT_BUTTON;
    break;
  case GLUT_MIDDLE_BUTTON:
    push_button = GLUT_MIDDLE_BUTTON;
    break;
  case GLUT_RIGHT_BUTTON:
    push_button = GLUT_RIGHT_BUTTON;
    break;
  }
  
  switch( state ){
  case GLUT_DOWN:
    break;
  case GLUT_UP:
    break;
  }

  calculationRatioOfMouseCoordinate( x, y );
}

/* ޥΥݥ󥿤ưä˸ƤФ */
static void
passiveMotionCallback( int x, int y )
{
}

/* ɥå桢ޥΥݥ󥿤ưä˸ƤФ */
static void
motionCallback( int x, int y )
{
  double rx, ry;

  switch( push_button ){ /* Ƥܥ˱ȿ򤵤 */
  case GLUT_LEFT_BUTTON:
    /* ʪΤβž׻ */
    rx = ratio_x;
    ry = ratio_y;
    calculationRatioOfMouseCoordinate( x, y );
    rotate_x += (float)((ratio_x-rx)*ROTATION_COEFFICIENT);
    rotate_y += (float)((ratio_y-ry)*ROTATION_COEFFICIENT);
    glutPostRedisplay();
    break;
  case GLUT_MIDDLE_BUTTON:
    /* ȥεΥѲ */
    rx = ratio_x;
    ry = ratio_y;
    calculationRatioOfMouseCoordinate( x, y );
    distance += ((float)(ratio_x-rx)+(float)(ratio_y-ry))*DISTANCE_COEFFICIENT;
    glutPostRedisplay();
    break;
  case GLUT_RIGHT_BUTTON:
    break;
  }
}

/* Ϥä˸ƤФ */
static void
keyboardCallback( unsigned char key_code, int x, int y )
{
  switch( key_code ){
  /*
  ** Esc줿ץλ
  */
  case 27:
    exit( EXIT_SUCCESS );
    break;

  case 'i':
    vy =  0.025f;
    break;

  case 'm':
    vy = -0.025f;
    break;

  case 'k':
    vx =  0.025f;
    break;

  case 'j':
    vx = -0.025f;
    break;

  default:
    break;
  }
}

/* OpenGL ν */
static void
initGL( void )
{
  int width, height;

  width  = glutGet( GLUT_WINDOW_WIDTH );
  height = glutGet( GLUT_WINDOW_HEIGHT );

  reshapeCallback( width, height );

  glClearColor( 0.0, 0.0, 0.0, 1.0 );
  glEnable( GL_DEPTH_TEST );
  glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
}

/* MAIN */
int
main( int argc, char *argv[] )
{
  glutInit( &argc, argv );
  glutInitDisplayMode( GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH );

  glutCreateWindow( "TIPS: Simple Collision Detection" );

  glutDisplayFunc      ( displayCallback       );
  glutKeyboardFunc     ( keyboardCallback      );
  glutReshapeFunc      ( reshapeCallback       );
  glutIdleFunc         ( idleCallback          );
  glutVisibilityFunc   ( visibilityCallback    );
  glutMouseFunc        ( mouseCallback         );
  glutPassiveMotionFunc( passiveMotionCallback );
  glutMotionFunc       ( motionCallback        );
  
  initGL(); /* OpenGL ν */
  initializeAll(); /*  */
  
  glutMainLoop(); /* ᥤ롼פγ */

  freeAll(); /*  */
  
  return EXIT_SUCCESS;
}
