/* trimodel.c	-- Yasuhiro Yamazaki Tue Dec 18 2001
 */
#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>

#include <GL/gl.h>

#include "vector3d.h"
#include "trimodel.h"

/* digit */
static int
digit( const char x )
{
  if( x>='0' && x<='9' ) return (int)x-(int)'0';
  if( x>='a' && x<='f' ) return (int)x-(int)'a' + 10;
  if( x>='A' && x<='F' ) return (int)x-(int)'A' + 10;
  return 0;
}

/* construct Triangles */
static void
constructTriangles( FILE *read )
{
  int i;
  Vector3D vector[3], v0, v1, normal;
  char index[256], index_reserve[256];
  int r, g, b;
  GLfloat color[4]; /* zero[4] = { 0.0f, 0.0f, 0.0f, 1.0f }; */
  color[3] = 1.0f;
  index_reserve[0] = '\0';

  glMaterialf ( GL_FRONT, GL_SHININESS, 32.0f );
    
  while( !feof(read) ){
    /* ǡɤ߹ */
    for( i=0; i<3; i++ ){
      fscanf( read, "%f%f%f", &vector[i].x, &vector[i].y, &vector[i].z );
    }
    fscanf( read, "%s", index );

    /* ̤Υޥƥꥢꤹ */
    if( strcmp( index, index_reserve )!=0 ){
      r = digit( index[2] ) * 16 + digit( index[3] );
      g = digit( index[4] ) * 16 + digit( index[5] );
      b = digit( index[6] ) * 16 + digit( index[7] );
      color[0] = (GLfloat)r/255.0f;
      color[1] = (GLfloat)g/255.0f;
      color[2] = (GLfloat)b/255.0f;
      
      glColor3f( color[0], color[1], color[2] );
      glMaterialfv( GL_FRONT, GL_DIFFUSE, color );
      glMaterialfv( GL_FRONT, GL_SPECULAR, color );
      glMaterialfv( GL_FRONT, GL_AMBIENT, color );

      strcpy( index_reserve, index );
    }
    
    /* ˡ׻ */
    subVector3D( &v0, &vector[1], &vector[0] );
    subVector3D( &v1, &vector[2], &vector[0] );
    crossProductVector3D( &normal, &v0, &v1 );
    normalizeVector3D( &normal );
    glNormal3f( normal.x, normal.y, normal.z );    

    /* ĺ󤹤 */
    for( i=0; i<3; i++ ){
      glVertex3f( vector[i].x, vector[i].y, vector[i].z );
    }
  }
}

/* Loader */
TriModel *
loadTriModel( const char *filename )
{
  TriModel *tri_model = NULL;
  FILE *read = fopen( filename, "r" );
  if( read!=NULL ){

    tri_model = (TriModel *)malloc( sizeof( TriModel ) );
    tri_model->ref = 1;
    tri_model->display_list = glGenLists( 1 );

    /* ǥץ쥤ꥹȤκ */
    glNewList( tri_model->display_list, GL_COMPILE );
    glBegin( GL_TRIANGLES );
    constructTriangles( read ); /* ѷ󤹤 */
    glEnd();
    glEndList();

    fclose( read );
  } else {
    /* ERROR */
    fprintf( stderr, "File \"%s\" not found.\n", filename );
  }
  return tri_model;
}

/* Put Model */
void 
putTriModel( const TriModel *tri_model )
{
  if( tri_model!=NULL && tri_model->display_list>0 ){
    /* ǥץ쥤ꥹȤθƤӽФ */
    glPushMatrix();
    glCallList( tri_model->display_list );
    glPopMatrix();
  }
}

/* Destroy */
void 
destroyTriModel( TriModel *tri_model )
{
  if( tri_model!=NULL ){
    if( --tri_model->ref<=0 ){
      /* ǡ˴ */
      if( tri_model->display_list>0 ){
        glDeleteLists( tri_model->display_list, 1 );
      }
      free( tri_model );
    }
  }
}
