/* vector3d.c	-- Yasuhiro Yamazaki Wed Oct 24 2001
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector3d.h"

/*  */
Vector3D *
createVector3D( void )
{
  Vector3D *ret = (Vector3D *)malloc( sizeof( Vector3D ) );
  ret->ref = 1;
  ret->x = 0.0f;
  ret->y = 0.0f;
  ret->z = 0.0f;
  return ret;
}

Vector3D *
createVector3Df( const float x, const float y, const float z )
{
  Vector3D *ret = createVector3D();
  setVector3DCoordinatef( ret, x, y, z );
  return ret;
}

Vector3D *
createVector3Dv( const float *v )
{
  Vector3D *ret = createVector3D();
  setVector3DCoordinatev( ret, v );
  return ret;
}

/*  */
Vector3D *
croneVector3D( const Vector3D *v )
{
  Vector3D *ret = createVector3D();
  copyVector3D( ret, v );
  return ret;
}

/* ˴ */
void 
destroyVector3D( Vector3D *v )
{
  if( v!=NULL ){
    if( --v->ref<=0 ){
      free( v );
    }
  }
}

/* ͤ */
void 
setVector3DCoordinatef( Vector3D *v, const float x, const float y, const float z )
{
  v->x = x;
  v->y = y;
  v->z = z;
}

void 
setVector3DCoordinatev( Vector3D *v, const float *f )
{
  v->x = *f++;
  v->y = *f++;
  v->z = *f;
}

void
copyVector3D( Vector3D *v, const Vector3D *from )
{
  v->x = from->x;
  v->y = from->y;
  v->z = from->z;
}

/* ͤη׻ */
void 
addVector3D( Vector3D *v, Vector3D *a, Vector3D *b ) /* v = a + b; */
{
  v->x = a->x + b->x;
  v->y = a->y + b->y;
  v->z = a->z + b->z;
}

void 
subVector3D( Vector3D *v, Vector3D *a, Vector3D *b ) /* v = a - b; */
{
  v->x = a->x - b->x;
  v->y = a->y - b->y;
  v->z = a->z - b->z;
}

void 
multiplyVector3D( Vector3D *v, const float a ) /* v *= a; */
{
  v->x *= a;
  v->y *= a;
  v->z *= a;
}

void 
divideVector3D( Vector3D *v, const float a ) /* v /= a; */
{
  float b = 1.0f/a;
  v->x *= b;
  v->y *= b;
  v->z *= b;
}

float
normalizeVector3D( Vector3D *v )
{
  float n;
  divideVector3D( v, (n=getNormal( v )) );
  return n;
}

float 
dotProductVector3D( Vector3D *a, Vector3D *b )
{
  return a->x * b->x + a->y * b->y + a->z * b->z;
}

void 
crossProductVector3D( Vector3D *v, Vector3D *a, Vector3D *b )
{
  float x, y, z;
  x = a->y * b->z - a->z * b->y;
  y = a->z * b->x - a->x * b->z;
  z = a->x * b->y - a->y * b->x;
  v->x = x;
  v->y = y;
  v->z = z;
}

float 
getNormal( Vector3D *v )
{
  float x = v->x, y = v->y, z = v->z;
  return (float)sqrt( x*x+y*y+z*z );
}
